@@ -1,4 +1,5 @@
 import os
+from typing import Optional
 import matplotlib.pyplot as plt
 plt.rcParams['figure.dpi']=200
 import numpy as np
@@ -130,4 +131,67 @@ def plot4d(func, z_values, path="", wbounds=None, frame=Frame2D(), save_images=F
         
     print("Animation saved as \"%s\""%gif_name)
     
-    return gif_name
\ No newline at end of file
+    return gif_name
+
+@dataclass
+class Frame3D: 
+    xmin: Optional[float] = None
+    xmax: Optional[float] = None
+    ymin: Optional[float] = None
+    ymax: Optional[float] = None
+    zmin: Optional[float] = None
+    zmax: Optional[float] = None
+    xlabel: str = "x"
+    ylabel: str = "y"
+    zlabel: str = "z"
+    
+def plot4d_data(X: np.ndarray, y: np.ndarray, frame=Frame3D(), best_highlight=False, title="test", c_label="", save=False):
+    """Plot 4 Dimensional Data. Optionally saves to "{title}.png"
+
+    Args:
+        X (np.ndarray): Data matrix. Shape (N, 3), where N is the number of data points
+        y (np.ndarray): Label matrix. Shape (N, )
+        frame (Frame3D, optional): 3D frame of the plot. Any limit not set in frame will be auto fitted. 
+        best_highlight (bool, optional): Highlight the point with largest y value. Defaults to False.
+        title (str, optional): Title of the plot. Defaults to "test".
+        c_label (str, optional): Colorbar label. Defaults to empty string.
+        save (bool, optional): Choose to save image. Defaults to False. 
+    """
+    assert X.shape[0] == y.shape[0]
+    
+    if len(y.shape) > 1:
+            # TODO: plot points with different sizes as the 5th dimension
+            raise NotImplementedError()
+    
+    xpts, ypts, zpts, sizepts, magpts = list(X[:,0]), list(X[:,1]), list(X[:,2]), list(np.zeros_like(X[:,0])), list(y)
+    
+    # plotting
+    fig = plt.figure(figsize=(6, 6))
+    ax = fig.add_subplot(111, projection='3d')
+
+    
+    xr, yr, zr, sr, mr = xpts[:],  ypts[:],  zpts[:], sizepts[:], magpts[:]
+    if best_highlight:
+        # seperate the best from the rest
+        bi = np.argmax(magpts) # best index
+        xb, yb, zb, sb, mb = xr.pop(bi),  yr.pop(bi),  zr.pop(bi), sr.pop(bi), mr.pop(bi)
+
+    graph = ax.scatter(xr, yr, zr,
+            c=mr, vmin=min(magpts),vmax=max(magpts), linewidths=0.5, alpha=.7)
+    if best_highlight:
+        ax.scatter(xb, yb, zb, 
+                c=mb, vmin=min(magpts),vmax=max(magpts), edgecolor='red', linewidths=0.5, alpha=.7)
+
+    cbar = fig.colorbar(graph)
+    cbar.set_label(c_label)
+    ax.set_xlabel(frame.xlabel)
+    ax.set_ylabel(frame.ylabel)
+    ax.set_zlabel(frame.zlabel)
+
+    if frame.xmin and frame.xmax: ax.set_xlim([frame.xmin,frame.xmax])
+    if frame.ymin and frame.ymax: ax.set_ylim([frame.ymin,frame.ymax])
+    if frame.zmin and frame.zmax: ax.set_zlim([frame.zmin,frame.zmax]) #TODO: (wishlist) enable user to set max and min seperately
+
+    plt.title(title)
+    if save: plt.savefig(title+".png")
+    plt.show()
\ No newline at end of file
